
#include "../../biscuit/biscuit.hpp"

namespace xml_grammar
{
	using namespace biscuit;

	struct Char :
		alternative<
			ucs4<0x9>, ucs4<0xA>, ucs4<0xD>,
			ucs4_range<0x20,0xD7FF>,
			ucs4_range<0xE000,0xFFFD>,
			ucs4_range<0x10000,0x10FFFF>
		>
	{ };

	struct CharData :
		kleene_star_before< Char, char_set<'<','&'> >
	{ };

	struct Comment :
		sequence<
			string<'<','!','-','-'>,
			kleene_star<
				alternative<
					difference< Char, string<'-'> >,
					sequence<
						string<'-'>,
						difference<	Char, string<'-'> >
					>
				>
			>,
			string<'-','-','>'>
		>
	{ };

	struct S :
		positive< char_set<0x20,0x9,0xD,0xA> >
	{ };

	struct Digit :
		alternative<
			ucs4_range<0x0030,0x0039>, ucs4_range<0x0660,0x0669>
			// too many
		>
	{ };

	struct BaseChar :
		alternative<
			ucs4_range<0x0041,0x005A>, ucs4_range<0x0061,0x007A>
		>
	{ };

	struct Ideographic :
		alternative<
			ucs4_range<0x4E00,0x9FA5>, ucs4<0x3007>,
			ucs4_range<0x3021,0x3029>
		>
	{ };

	struct CombiningChar :
		alternative<
			ucs4_range<0x0300,0x0345>, ucs4_range<0x0360,0x0361>
			// too many
		>
	{ };

	struct Extender :
		alternative<
			ucs4<0x00B7>, ucs4<0x02D0>, ucs4<0x02D1>, ucs4<0x0387>
			// many
		>
	{ };

	struct Letter :
		alternative< BaseChar, Ideographic >
	{ };

	struct NameChar :
		alternative<
			Letter, Digit, string<'.'>, string<'-'>, string<'_'>, string<':'>,
			CombiningChar, Extender
		>
	{ };

	struct Name :
		sequence<
			alternative< Letter, string<'_'>, string<':'> >,
			kleene_star< NameChar >
		>
	{ };

	struct Eq :
		sequence< optional<S>, string<'='>, optional<S> >
	{ };

	struct PITarget :
		difference<
			Name,
			sequence<
				alternative< string<'X'>, string<'x'> >,
				alternative< string<'M'>, string<'m'> >,
				alternative< string<'L'>, string<'l'> >
			>
		>
	{ };
		
	struct PI :
		sequence<
			string<'<','?'>,
			PITarget,
			optional<
				sequence< S, kleene_star_before< Char, string<'?','>'> > >
			>,
			string<'?','>'>
		>
	{ };

	struct CDStart :
		string<'<','!','[','C','D','A','T','A','['>
	{ };

	struct CData :
		kleene_star_before< Char, string<']',']','>'> >
	{ };

	struct CDEnd :
		string<']',']','>'>
	{ };

	struct CDSect :
		sequence< CDStart, CData, CDEnd >
	{ };

	struct doctypedecl :
		sequence<
			string<'<','!','D','O','C','T','Y','P','E'>,
			kleene_star_until< Char, string<'>'> >
		>
	{ };

	struct XMLDecl :
		sequence<
			string<'<','?','x','m','l'>,
			kleene_star_until< Char, string<'?','>'> >
		>
	{ };

	struct Misc :
		alternative< Comment, PI, S >
	{ };

	struct prolog :
		sequence<
			optional<XMLDecl>,
			kleene_star<Misc>,
			optional< sequence< doctypedecl, kleene_star<Misc> > >
		>
	{ };

	struct AttValue :
		alternative<
			sequence<
				string<'"'>,
				kleene_star_until< Char, string<'"'> >
			>,
			sequence<
				string<'\''>,
				kleene_star_until< Char, string<'\''> >
			>
		>
	{ };

	struct Attribute :
		sequence< Name, Eq, AttValue >
	{ };

	struct EmptyElemTag :
		sequence<
			string<'<'>,
			Name,
			kleene_star< sequence<S,Attribute> >,
			optional<S>,
			string<'/','>'>
		>
	{ };

	struct STag :
		sequence<
			string<'<'>,
			Name,
			kleene_star< sequence<S,Attribute> >,
			optional<S>,
			string<'>'>
		>
	{ };

	struct ETag :
		sequence<
			string<'<','/'>,
			Name,
			optional<S>,
			string<'>'>
		>
	{ };

	struct CharRef :
		alternative<
			sequence<
				string<'&','#'>,
				positive< char_range<'0','9'> >,
				string<';'>
			>,
			sequence<
				string<'&','#','x'>,
				positive< alternative< char_range<'0','9'>, char_range<'a','f'>, char_range<'A','F'> > >,
				string<';'>
			>
		>
	{ };

	struct EntityRef :
		sequence< string<'&'>, Name, string<';'> >
	{ };

	struct Reference :
		alternative<EntityRef, CharRef>
	{ };

	
	struct element;

	struct content :
		sequence<
			optional<CharData>,
			kleene_star<
				sequence<
					alternative<
						element, // the magical recursion!
						Reference,
						CDSect,
						PI,
						Comment
					>,
					optional<CharData>
				>
			>
		>
	{ };

	struct element :
		alternative<
			EmptyElemTag, 
			sequence<STag, content, ETag>
		>
	{ };

	struct document :
		sequence<
			prolog,
			element,
			kleene_star<Misc>
		>
	{ };

} // namespace xml_grammar
